import contextlib
import queue
import pickle
import os
import sys
from rich import print
from histdatacom.fx_enums import Timeframe, Format
from histdatacom.utils import get_month_from_datemonth
from histdatacom.utils import get_year_from_datemonth
from histdatacom.utils import get_query_string
from histdatacom.utils import create_full_path


class Record:
    def __init__(self, **kwargs):
        self.url = kwargs.get("url", "")
        self.status = kwargs.get("status", "")
        self.encoding = kwargs.get("encoding", "")
        self.bytes_length = kwargs.get("bytes_length", "")
        self.data_date = kwargs.get("data_date", "")
        self.data_year = kwargs.get("data_year", "")
        self.data_month = kwargs.get("data_month", "")
        self.data_datemonth = kwargs.get("data_datemonth", "")
        self.data_format = kwargs.get("data_format", "")
        self.data_timeframe = kwargs.get("data_timeframe", "")
        self.data_fxpair = kwargs.get("data_fxpair", "")
        self.data_dir = kwargs.get("data_dir", "")
        self.data_tk = kwargs.get("data_tk", "")
        self.zip_filename = kwargs.get("zip_filename", "")
        self.csv_filename = kwargs.get("csv_filename", "")
        self.jay_filename = kwargs.get("jay_filename", "")
        self.csv_linecount = kwargs.get("csv_linecount", "")
        self.jay_linecount = kwargs.get("jay_linecount", "")
        self.jay_start = kwargs.get("jay_start", "")
        self.jay_end = kwargs.get("jay_end", "")
        self.zip_persist = kwargs.get("zip_persist", "")

    def __call__(self, string="updated", **kwargs):
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])
        return self

    def set_datetime_attrs(self):
        self.data_year = get_year_from_datemonth(self.data_datemonth)
        self.data_month = get_month_from_datemonth(self.data_datemonth)

    def set_record_data_dir(self, base_dir):
        query_string_args = get_query_string(self.url)
        length = len(query_string_args)

        csv_format = Format(query_string_args[1]).name
        timeframe = Timeframe(query_string_args[2]).name

        record_data_dir = base_dir + csv_format + os.sep + timeframe + os.sep

        if length == 3:
            self.data_dir = record_data_dir
            return self.data_dir

        pair = query_string_args[3]
        record_data_dir = record_data_dir + pair.lower() + os.sep

        if length == 4:
            self.data_dir = record_data_dir
            return self.data_dir

        year = query_string_args[4]
        record_data_dir = record_data_dir + year + os.sep

        if length == 5:
            self.data_dir = record_data_dir
            return self.data_dir

        month = query_string_args[5]
        record_data_dir = record_data_dir + month + os.sep

        if length == 6:
            self.data_dir = record_data_dir
            return self.data_dir

    def set_csv_linecount(self):
        with open(f"{self.data_dir}{self.csv_filename}") as csv:
            self.csv_linecount = sum(1 for _ in csv)

    def create_record_data_dir(self, base_dir=""):
        try:
            if self.data_dir != "":
                create_full_path(self.data_dir)
            elif base_dir != "":
                create_full_path(self.set_record_data_dir(base_dir))
            else:
                raise ValueError("Error: create_record_data_dir not provided base_dir=")
        except ValueError as err:
            print(err)
            sys.exit()

    def write_info_file(self, base_dir=""):
        try:
            if self.data_dir == "":
                if base_dir != "":
                    self.create_record_data_dir(base_dir=base_dir)
                else:
                    raise ValueError("Error: create_record_data_dir not provided base_dir=")

            if not os.path.exists(self.data_dir):
                create_full_path(self.data_dir)

            path = f'{self.data_dir}.info'

            with open(path, 'wb') as filepath:
                pickle.dump(self.to_dict(), filepath)
        except ValueError as err:
            print(err)
            sys.exit()

    def delete_info_file(self):
        if os.path.exists(f"{self.data_dir}.info"):
            os.remove(f"{self.data_dir}.info")

    def restore_momento(self, base_dir):
        self.set_record_data_dir(base_dir)
        if not os.path.exists(f"{self.data_dir}.info"):
            return False
        record_dict = {}
        with open(f"{self.data_dir}.info", 'rb') as fileread:
            with contextlib.suppress(Exception):
                while True:
                    record_dict.update(pickle.load(fileread))
        self(**record_dict)
        return True

    def to_dict(self):
        return {'url': self.url,
                'status': self.status,
                'encoding': self.encoding,
                'bytes_length': self.bytes_length,
                'data_date': self.data_date,
                'data_year': self.data_year,
                'data_month': self.data_month,
                'data_datemonth': self.data_datemonth,
                'data_format': self.data_format,
                'data_timeframe': self.data_timeframe,
                'data_fxpair': self.data_fxpair,
                'data_dir': self.data_dir,
                'data_tk': self.data_tk,
                'zip_filename': self.zip_filename,
                'csv_filename': self.csv_filename,
                'csv_linecount': self.csv_linecount,
                'jay_linecount': self.jay_linecount,
                'jay_start': self.jay_start,
                'jay_end': self.jay_end,
                'jay_filename': self.jay_filename,
                'zip_persist': self.zip_persist}

    def print_record(self, string="Updated"):
        print(f"{string}:",
              self.status,
              self.data_format,
              self.data_timeframe,
              self.data_fxpair,
              self.data_year,
              self.data_month,
              "-",
              self.data_dir)


class Records(queue.Queue):
    def __init__(self, *args, **kwargs):
        queue.Queue.__init__(self, *args, **kwargs)

    def print(self):
        printable = []

        while not self.empty():
            record = self.get()
            if record is None:
                break

            printable.append(record.to_dict())

        for p_record in printable:
            new_record = Record(**p_record)
            print(new_record.to_dict())
            self.put(new_record)

    def write_pickle(self, path):
        picklable = []

        while not self.empty():
            record = self.get()
            if record is None:
                break

            picklable.append(record.to_dict())

        with open(path, 'wb') as filepath:
            pickle.dump(picklable, filepath)

        for p_record in picklable:
            new_record = Record(**p_record)
            self.put(new_record)

    def __contains__(self, item):
        with self.mutex:
            return item in self.queue

    def __len__(self):
        return len(self.queue)

    def dump_to_queue(self, dst_queue, count=0):
        if count == 0:
            _count = None
        else:
            _count = count
            _counter = 0

        while not self.empty():
            record = self.get()

            if record is None:
                break

            if _count is None:
                dst_queue.put(record)
            elif _counter < _count:
                dst_queue.put(record)
                _counter += 1
            else:
                self.put(record)
                break



"""_summary_
CLI interface for histdatacom

Raises:
    ValueError: ERROR on -f {args_namespace.formats}           ERROR
                    * format must be ASCII when importing to influxdb. eg. -f ascii
    ValueError: ERROR on -e start           ERROR
                                    * keyword 'start' cannot be used as -e start
    ValueError: ERROR on -s start           ERROR
                                    * keyword 'start' must also specify
                                      an end year-month
    ValueError: ERROR on -e {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` without
                          specifying a start year-month.
                            eg. -s year-month -e year-month
    ValueError: ERROR on -e {get_year_from_datemonth(start_yearmonth)}  {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` when
                          specifying a single year
                          with -s {get_year_from_datemonth(start_yearmonth)}
    ValueError: ERROR on -e {start_yearmonth}   ERROR
                    start month cannot be zero.
                        * valid inputs:
                            a) just the year
                                eg. -s 2022
                            b) months 1-12:
                                eg. -s 2022-04
    ValueError: ERROR on -s {start_yearmonth}  ERROR
                    * Malformed command:
                        - start month is greater than 12.
                          valid input is 01-12.
    ValueError: ERROR on -e {end_yearmonth}           ERROR
                            * You left out the end month.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03
    ValueError: ERROR on -e {end_yearmonth}           ERROR
                            * End month cannot be zero.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03
    ValueError: ERROR on -e {end_yearmonth}  ERROR
                            * Malformed command:
                                - end month is greater than 12.
                                valid input is 01-12.
    ValueError: ERROR on -s {start_yearmonth} -e {end_yearmonth}  ERROR
                    * start year-month and end year-month cannot be the same.
    ValueError: ERROR on -s {start_yearmonth}      ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
    ValueError: ERROR on -s {start_yearmonth}      ERROR
                            * year-month cannot be in the future.
    ValueError: ERROR on -e {end_yearmonth}     ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
    ValueError: ERROR on -e {end_yearmonth}     ERROR
                            * year-month cannot be in the future.
    ValueError: ERROR on -s {start_yearmonth} -e {end_yearmonth}    ERROR
                            * logic error: end year-month is before start year-month.
    ValueError: ERROR on {yearmonth}    ERROR
                            * invalid yearmonth format

Returns:
    _type_: self.arg_namespace
"""

import argparse
import sys
import re
from histdatacom.fx_enums import Pairs
from histdatacom.fx_enums import Format
from histdatacom.fx_enums import Timeframe
from histdatacom.utils import get_current_datemonth_gmt_plus5
from histdatacom.utils import get_month_from_datemonth
from histdatacom.utils import get_year_from_datemonth
from histdatacom.utils import replace_date_punct
from histdatacom.options import Options


class ArgParser(argparse.ArgumentParser):
    """ Encapsulation class for argparse related operations """

    def __init__(self, options=Options(), **kwargs):
        """ set up argparse, bring in defaults DTO, setup cli params, receive
            and overwrite defaults with user cli args."""
        
        # init _HistDataCom.ArgParser to extend argparse.ArgumentParser
        argparse.ArgumentParser.__init__(self, prog='histdatacom')
        # bring in the defaults arg DTO from outer class, use the
        # __dict__ representation of it to set argparse argument defaults.
        self.arg_namespace = options
        self._default_args = self.arg_namespace.__dict__
        self.set_defaults(**self._default_args)

        # Nothing special here, adding cli params
        # metavar="..." is used to limit the display of choices="large iterables".
        self.add_argument(
            "-V", "--validate_urls",
            action='store_true',
            help='Check generated list of URLs as valid download locations')
        self.add_argument(
            "-D", "--download_data_archives",
            action='store_true',
            help='download specified pairs/formats/timeframe and create data files')
        self.add_argument(
            "-X", "--extract_csvs",
            action='store_true',
            help='histdata.com delivers zip files.  use the -X flag to extract them to .csv.')
        self.add_argument(
            "-I", "--import_to_influxdb",
            action='store_true',
            help='import csv data to influxdb instance. Use influxdb.yaml to configure.')
        self.add_argument(
            '-p', '--pairs',
            nargs='+',
            type=str,
            choices=Pairs.list_keys(),
            help='space separated currency pairs. e.g. -p eurusd usdjpy ...',
            metavar='PAIR')
        self.add_argument(
            '-f', '--formats',
            nargs='+',
            type=str,
            choices=Format.list_values(),
            help='space separated formats. e.g. -f metatrader ascii ninjatrader metastock',
            metavar='FORMAT')
        self.add_argument(
            '-t', '--timeframes',
            nargs='+',
            type=(lambda v: Timeframe(v).name),  # convert long Timeframe .value to short .key
            choices=Timeframe.list_keys(),
            help='space separated Timeframes. e.g. -t tick-data-quotes 1-minute-bar-quotes ...',
            metavar='TIMEFRAME')
        self.add_argument(
            "-s", "--start_yearmonth",
            type=(lambda v: self.validate_yearmonth_format(v)),
            help='set a start year and month for data. e.g. -s 2000-04 or -s 2015-00')
        self.add_argument(
            "-e", "--end_yearmonth",
            type=(lambda v: self.validate_yearmonth_format(v)),
            help='set a start year and month for data. e.g. -s 2020-00 or -s 2022-04')
        self.add_argument(
            '-d', '--data-directory',
            type=str,
            help='Directory Used to save data. default is "data" in the current directory')
        
        if "histdatacom" not in sys.argv[0] and self.arg_namespace.from_api:
            args = self.clean_from_api_args(self.arg_namespace)
            self.false_from_api_if_behavior_flag(self.arg_namespace)
            self.parse_args(args, namespace=self.arg_namespace)
        else:
            # Get the args from sys.argv
            self.parse_args(namespace=self.arg_namespace)

        # prevent running from cli with no arguments
        if len(sys.argv) == 1 and not self.arg_namespace.from_api:
            self.print_help(sys.stderr)
            sys.exit(1)

        self.check_datetime_input(self.arg_namespace)
        self.check_for_ascii_if_influx(self.arg_namespace)
        self.check_for_ascii_if_api(self.arg_namespace)



    def __call__(self):
        """ simply return the completed args object """
        return self.arg_namespace

    @classmethod
    def clean_from_api_args(cls, args_namespace):
        args = []
        
        args.extend(["-d", args_namespace.data_directory])
        args.extend(["-p", *args_namespace.pairs])
        args.extend(["-f", *args_namespace.formats])
        args.extend(["-t", *args_namespace.timeframes])

        if args_namespace.start_yearmonth:
            args.extend(["-s", args_namespace.start_yearmonth])
        if args_namespace.end_yearmonth:
            args.extend(["-e", args_namespace.end_yearmonth])
        if args_namespace.validate_urls:
            args.append("-V")
        if args_namespace.download_data_archives:
            args.append("-D")
        if args_namespace.extract_csvs:
            args.append("-X")
        if args_namespace.import_to_influxdb:
            args.append("-I")
        return args

    @classmethod
    def _arg_list_to_set(cls, args):
        # This is to standardize data types. If the user specifies a parameter,
        # argparse returns a list, our defaults are sets, so .
        for arg in args:
            if isinstance(args[arg], list):
                args[arg] = set(args[arg])
        return args
    @classmethod
    def false_from_api_if_behavior_flag(cls, args_namespace):
        if args_namespace.validate_urls \
        or args_namespace.download_data_archives \
        or args_namespace.extract_csvs \
        or args_namespace.import_to_influxdb:
            args_namespace.from_api = False
            
    @classmethod
    def check_for_ascii_if_influx(cls, args_namespace):
        """Verify ascii csv_format type for influxdb import"""
        try:
            err_text_influx_must_be_ascii = \
            f"""
                ERROR on -f {args_namespace.formats}           ERROR
                    * format must be ASCII when importing to influxdb. eg. -f ascii

            """
            if args_namespace.import_to_influxdb:
                for csv_format in args_namespace.formats:
                    if str.lower(csv_format) != "ascii":
                        raise ValueError(err_text_influx_must_be_ascii)
        except ValueError as err:
            print(err)
            sys.exit(err)

    @classmethod
    def check_for_ascii_if_api(cls, args_namespace):
        try:
            err_text_api_must_be_ascii = \
            f"""
                ERROR on -f {args_namespace.formats}           ERROR
                    * format must be ASCII when calling from API 
                        eg. 
                            import histdatacom
                            from histdatacom.options import Options

                            options = Options()
                            options.formats = {{"ascii"}}
            """
            if args_namespace.from_api \
            and not (
                args_namespace.validate_urls 
                or args_namespace.download_data_archives \
                or args_namespace.extract_csvs \
                or args_namespace.import_to_influxdb
            ):
                for csv_format in args_namespace.formats:
                    if str.lower(csv_format) != "ascii":
                        raise ValueError(err_text_api_must_be_ascii)
        except ValueError as err:
            print(err)
            sys.exit()
    
    
    @classmethod
    def check_datetime_input(cls, args_namespace):
        """Checks for invalid datetime input for -s and -e flags"""
        if args_namespace.start_yearmonth \
        or args_namespace.end_yearmonth:
            args_namespace.start_yearmonth, args_namespace.end_yearmonth = \
                cls.check_for_start_in_yearmonth(args_namespace)

            args_namespace.start_yearmonth, args_namespace.end_yearmonth = \
                cls.check_for_now_in_yearmonth(args_namespace)

            args_namespace.start_yearmonth, args_namespace.end_yearmonth = \
                cls.check_cli_start_yearmonth(args_namespace)

            cls.check_cli_end_yearmonth(args_namespace)

            cls.check_for_same_start_yearmonth(args_namespace)

        args_namespace.start_yearmonth, args_namespace.end_yearmonth = \
            cls.replace_falsey_yearmonth_with_none(args_namespace)

        cls.check_start_yearmonth_in_range(args_namespace)
        cls.check_end_yearmonth_in_range(args_namespace)
        cls.check_start_lessthan_end(args_namespace)
        cls.validate_prerequisites(args_namespace)

    @classmethod
    def validate_prerequisites(cls, args_namespace):
        """Sets prereqs for behavior flags -V -D -X -I"""
        if args_namespace.validate_urls:
            return

        if args_namespace.download_data_archives:
            args_namespace.validate_urls = True

        if args_namespace.extract_csvs:
            args_namespace.validate_urls = True
            args_namespace.download_data_archives = True

        if args_namespace.import_to_influxdb:
            args_namespace.validate_urls = True
            args_namespace.download_data_archives = True
            args_namespace.extract_csvs = True

        if not args_namespace.download_data_archives \
        and not args_namespace.extract_csvs \
        and not args_namespace.import_to_influxdb:
            args_namespace.validate_urls = True
            args_namespace.download_data_archives = True
            args_namespace.extract_csvs = True

    @classmethod
    def check_for_now_in_yearmonth(cls, args_namespace):
        """checks for now in -s or -e and adjusts it to current year-month"""
        if (start_yearmonth := args_namespace.start_yearmonth):
            if start_yearmonth == "now":
                return get_current_datemonth_gmt_plus5(), None
            if end_yearmonth := args_namespace.end_yearmonth:
                if end_yearmonth == "now":
                    return start_yearmonth, get_current_datemonth_gmt_plus5()

        return start_yearmonth, end_yearmonth

    @classmethod
    def check_for_start_in_yearmonth(cls, args_namespace):
        """Checks for 'start' keyword in -s and sets -s yearmonth to 200001"""
        try:
            if (start_yearmonth := args_namespace.start_yearmonth) and (start_yearmonth == "start"):
                if (end_yearmonth := args_namespace.end_yearmonth):
                    if end_yearmonth == "start":
                        err_text_end_yearmonth_cannot_be_start = \
                        """
                            ERROR on -e start           ERROR
                                * keyword 'start' cannot be used as -e start
                        """
                        raise ValueError(err_text_end_yearmonth_cannot_be_start)
                    return "200001", end_yearmonth

                err_text_start_must_have_end = \
                """
                        ERROR on -s start           ERROR
                            * keyword 'start' must also specify
                                an end year-month
                """
                raise ValueError(err_text_start_must_have_end)
            return args_namespace.start_yearmonth, args_namespace.end_yearmonth
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def check_cli_start_yearmonth(cls, args_namespace):
        """Validations for -s start_yearmonth"""
        start_yearmonth = args_namespace.start_yearmonth
        start_year = get_year_from_datemonth(start_yearmonth)
        start_month = get_month_from_datemonth(start_yearmonth)

        err_text_start_month = \
        f"""
                ERROR on -e {start_yearmonth}   ERROR
                    start month cannot be zero.
                        * valid inputs:
                            a) just the year
                                eg. -s 2022
                            b) months 1-12:
                                eg. -s 2022-04
        """

        end_yearmonth = args_namespace.end_yearmonth

        err_text_no_end_yearmonth = \
        f"""
                ERROR on -e {get_year_from_datemonth(start_yearmonth)}  {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` when
                          specifying a single year
                          with -s {get_year_from_datemonth(start_yearmonth)}
        """

        err_text_no_start_yearmonth = \
        f"""
                ERROR on -e {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` without
                          specifying a start year-month.
                            eg. -s year-month -e year-month
        """
        err_text_start_month_greater_than_12 = \
        f"""
                ERROR on -s {start_yearmonth}  ERROR
                    * Malformed command:
                        - start month is greater than 12.
                          valid input is 01-12.
        """

        try:
            if not start_month:
                if end_yearmonth:
                    if not start_year:
                        raise ValueError(err_text_no_start_yearmonth)
                    raise ValueError(err_text_no_end_yearmonth)
                return f"{start_year}00", None
            if start_month == "00":
                raise ValueError(err_text_start_month)
            if int(start_month) > 12:
                raise ValueError(err_text_start_month_greater_than_12)
            return start_yearmonth, end_yearmonth
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def check_cli_end_yearmonth(cls, args_namespace):
        """Validations for -e end_yearmonth"""
        try:
            if end_yearmonth := args_namespace.end_yearmonth:
                end_year = get_year_from_datemonth(end_yearmonth)
                end_month = get_month_from_datemonth(end_yearmonth)

                err_text_no_endmonth = \
                f"""
                        ERROR on -e {end_yearmonth}           ERROR
                            * You left out the end month.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03

                """

                err_text_endmonth_cannot_be_zero = \
                f"""
                        ERROR on -e {end_yearmonth}           ERROR
                            * End month cannot be zero.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03

                """
                err_text_end_month_greater_than_12 = \
                f"""
                        ERROR on -e {end_yearmonth}  ERROR
                            * Malformed command:
                                - end month is greater than 12.
                                valid input is 01-12.
                """
                if end_year and not end_month:
                    raise ValueError(err_text_no_endmonth)
                if end_month == "00":
                    raise ValueError(err_text_endmonth_cannot_be_zero)
                if int(end_month) > 12:
                    raise ValueError(err_text_end_month_greater_than_12)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def check_for_same_start_yearmonth(cls, args_namespace):
        """Validates that -s start_yearmonth and -e end_yearmonth are not the same"""
        try:
            start_yearmonth = args_namespace.start_yearmonth
            start_year = get_year_from_datemonth(start_yearmonth)
            start_month = get_month_from_datemonth(start_yearmonth)

            end_yearmonth = args_namespace.end_yearmonth
            end_year = get_year_from_datemonth(end_yearmonth)
            end_month = get_month_from_datemonth(end_yearmonth)

            err_text_start_and_end_cannot_be_the_same = \
            f"""
                ERROR on -s {start_yearmonth} -e {end_yearmonth}  ERROR
                    * start year-month and end year-month cannot be the same.
            """

            if f"{start_year}_{start_month}" == f"{end_year}_{end_month}":
                raise ValueError(err_text_start_and_end_cannot_be_the_same)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def replace_falsey_yearmonth_with_none(cls, args_namespace):
        """if no user input on -s or -e set value to None"""
        start_yearmonth = args_namespace.start_yearmonth
        end_yearmonth = args_namespace.end_yearmonth

        if not start_yearmonth or start_yearmonth == "":
            start_yearmonth = None
        if not end_yearmonth or end_yearmonth == "":
            end_yearmonth = None

        return start_yearmonth, end_yearmonth

    @classmethod
    def check_start_yearmonth_in_range(cls, args_namespace):
        """Validate that -s is not earlier than 2000-01 or later than the current year-month"""
        try:
            if start_yearmonth := args_namespace.start_yearmonth:
                err_text_date_prior_to_dataset = \
                f"""
                        ERROR on -s {start_yearmonth}      ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
                """
                err_text_date_is_in_future = \
                f"""
                        ERROR on -s {start_yearmonth}      ERROR
                            * year-month cannot be in the future.
                """
                if int(start_yearmonth) < 200000:
                    raise ValueError(err_text_date_prior_to_dataset)
                if int(start_yearmonth) > int(get_current_datemonth_gmt_plus5()):
                    raise ValueError(err_text_date_is_in_future)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def check_end_yearmonth_in_range(cls, args_namespace):
        """checks if -e is not earlier than 200001 or later than current year-month"""
        try:
            if end_yearmonth := args_namespace.end_yearmonth:
                err_text_date_prior_to_dataset = \
                f"""
                        ERROR on -e {end_yearmonth}     ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
                """
                err_text_date_is_in_future = \
                f"""
                        ERROR on -e {end_yearmonth}     ERROR
                            * year-month cannot be in the future.
                """
                if int(end_yearmonth) < 200000:
                    raise ValueError(err_text_date_prior_to_dataset)
                if int(end_yearmonth) > int(get_current_datemonth_gmt_plus5()):
                    raise ValueError(err_text_date_is_in_future)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def check_start_lessthan_end(cls, args_namespace):
        """validats that -e is not a year-month earlier than -s"""
        try:
            if (start_yearmonth := args_namespace.start_yearmonth) \
            and (end_yearmonth := args_namespace.end_yearmonth):

                err_text_start_date_after_end_date = \
                    f"""
                        ERROR on -s {start_yearmonth} -e {end_yearmonth}    ERROR
                            * logic error: end year-month is before start year-month.
                    """
                if int(start_yearmonth) > int(end_yearmonth):
                    raise ValueError(err_text_start_date_after_end_date)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def validate_yearmonth_format(cls, yearmonth):
        """Initial user input validation.
            checks year-month format for:
                -  0000
                -  0000-00
                - "0000 00"
                -  0000_00
                -  0000.00
                -  0000:00
                -  000000
                -  start
                -  now

            returns str("000000") or "now" or "start"
        """
        try:
            err_text_bad_yearmonth_format = \
            f"""
                        ERROR on {yearmonth}    ERROR
                            * invalid yearmonth format
            """

            if re.match("^\d{4}[-_.: ]\d{2}$", yearmonth) \
            or re.match("^\d{6}$", yearmonth) \
            or re.match("^\d{4}$", yearmonth) \
            or str.lower(yearmonth) == "now" \
            or str.lower(yearmonth) == "start" \
            or yearmonth == "":
                return replace_date_punct(yearmonth)
            raise ValueError(err_text_bad_yearmonth_format)
        except ValueError as err:
            cls.exit_on_datetime_error(err)

    @classmethod
    def exit_on_datetime_error(cls, err):
        """standared exit on error for user input"""
        print(err)
        sys.exit(err)



import itertools
import sys
import os
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
import datatable as dt
from datatable import f
dt.options.progress.enabled = False
from rich.progress import Progress
from rich.progress import BarColumn
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from histdatacom.utils import replace_date_punct


class _API():
    def __init__(self, args_, records_current_, records_next_):
        # setting relationship to global outer parent
        self.args = args_

        global records_current
        records_current = records_current_

        global records_next
        records_next = records_next_

    def init_counters(self, records_current_, records_next_, args_):
        global records_current
        records_current = records_current_
        global records_next
        records_next = records_next_
        global args
        args = args_

    def create_jay(self, record):
        zip_path = record.data_dir + record.zip_filename
        zip_data = self.import_file_to_datatable(record, zip_path)

        record.jay_filename = ".data"
        jay_path = record.data_dir + record.jay_filename
        self.export_datatable_to_jay(zip_data, jay_path)

        record.jay_linecount = zip_data.nrows
        record.jay_start = self.extract_single_value_from_frame(zip_data, 0, "datetime")
        record.jay_end = self.extract_single_value_from_frame(zip_data, zip_data.nrows - 1, "datetime")
        record.write_info_file(base_dir=self.args['default_download_dir'])

    def validate_jay(self, record):
        try:
            if str.lower(record.data_format) == "ascii" \
            and (record.data_timeframe == "T"
                 or record.data_timeframe == "M1"):
                if "CSV" in record.status:
                    self.create_jay(record)

            records_next.put(record, self.args)
        except Exception:
                print("Unexpected error:", sys.exc_info())
                record.delete_info_file()
                raise
        finally:
                records_current.task_done()

    def validate_jays(self, records_current, records_next):
            
            records_count = records_current.qsize()
            with Progress(TextColumn(text_format=f"[cyan]Staging {records_count} datafiles..."),
                          BarColumn(),
                          "[progress.percentage]{task.percentage:>3.0f}%",
                          TimeElapsedColumn()) as progress:
                task_id = progress.add_task("extract", total=records_count)
                
                with ProcessPoolExecutor(max_workers=(multiprocessing.cpu_count() - 1),
                                         initializer=self.init_counters, 
                                         initargs=(records_current,
                                                   records_next,
                                                   self.args.copy())) as executor:
                    futures = []

                    while not records_current.empty():
                        record = records_current.get()

                        if record is None:
                            return

                        future = executor.submit(self.validate_jay, record)
                        progress.advance(task_id, 0.25)
                        futures.append(future)
                    
                    for future in as_completed(futures):
                        progress.advance(task_id, 0.75)
                        futures.remove(future)
                        del future

            records_current.join()
            records_next.dump_to_queue(records_current)

    def merge_jays(self, records_current, records_next):

        records_to_merge = []
        pairs = []
        timeframes = []
        while not records_current.empty():
            record = records_current.get()

            if record is None:
                return

            if (record.jay_filename == ".data" 
            and os.path.exists(record.data_dir + record.jay_filename)):
                pairs.append(record.data_fxpair)
                timeframes.append(record.data_timeframe)
                records_to_merge.append(record)

        sets_to_merge = []
        for timeframe, pair in itertools.product(set(timeframes), set(pairs)):
            tp_set_dict = {
                'timeframe': timeframe,
                'pair': pair,
                'records': [],
                'data': None
            }
            for m_record in records_to_merge:
                if m_record.data_timeframe == timeframe \
                and m_record.data_fxpair == pair:
                    tp_set_dict['records'].append(m_record)
            sets_to_merge.append(tp_set_dict)
        

        for tp_set in sets_to_merge:
            self.merge_records(tp_set)

        if len(sets_to_merge) == 1:
            return sets_to_merge[0]["data"]
        return sets_to_merge

    def merge_records(self, tp_set_dict):
        match tp_set_dict['timeframe']:
            case "T":
                merged = dt.Frame(names=["datetime", "bid", "ask", "vol"])
            case "M1":
                merged = dt.Frame(names=["datetime", "open", "high", "low", "close", "vol"])

        tp_set_dict['records'].sort(key=lambda record: record.jay_start)

        records_count = len(tp_set_dict)
        with Progress(TextColumn(text_format=f"[cyan]Merging {records_count} records..."),
                        BarColumn(),
                        "[progress.percentage]{task.percentage:>3.0f}%",
                        TimeElapsedColumn()) as progress:
            task_id = progress.add_task("extract", total=records_count)

            for m_record in tp_set_dict['records']:
                jay_path = m_record.data_dir + m_record.jay_filename
                jay_data = self.import_jay_data(jay_path)
                merged.rbind(jay_data)

            if self.args['api_return_type'] == "datatable":
                tp_set_dict['data'] = merged
            if self.args['api_return_type'] == "arrow":
                tp_set_dict['data'] = merged.to_arrow()
            if self.args['api_return_type'] == "pandas":
                tp_set_dict['data'] = merged.to_pandas()
        
    @classmethod
    def extract_single_value_from_frame(cls, frame, row, column):
        return int(replace_date_punct(frame[row, column]))

    @classmethod
    def import_file_to_datatable(cls, record, zip_path):
        try:
            match record.data_timeframe:
                case "M1":
                    data = dt.fread(zip_path, 
                                    header=False,
                                    columns=["datetime", "open", "high", "low", "close", "vol"],
                                    multiple_sources="ignore")
                case "T":
                    data = dt.fread(zip_path,
                                    header=False,
                                    columns=["datetime", "bid", "ask", "vol"],
                                    multiple_sources="ignore")
                case _:
                    raise ValueError("Error creating jay")

            data['vol'] = dt.int32
            return data
        except ValueError as err:
            print(err)
            sys.exit(err)

    @classmethod
    def export_datatable_to_jay(cls, data_frame, file_path):
        data_path = file_path
        data_frame.to_jay(data_path)
        return 0

    @classmethod
    def import_jay_data(cls, jay_path):
        return dt.fread(jay_path)



import sys
from . import histdata_com

__version__ = "0.75.9"
__author__ = 'David Midlo'

class Options(sys.modules[__name__].__class__):
    def __call__(self, options):
        return histdata_com.main(options)


sys.modules[__name__].__class__ = Options


from enum import Enum
from influxdb_client import WritePrecision


# Majors 7
# eurusd usdjpy gbpusd usdcad usdchf audusd nzdusd

# Minors 7
# eurgbp euraud gbpchf audnzd audcad audchf gbpaud usdmxn

# Crosses 10
# eurchf eurcad eurnzd eurjpy gbpjpy chfjpy cadjpy
# audjpy nzdjpy gbpcad nzdcad sgdjpy gbpnzd cadchf

# Exotics 7
# eurtry usdtry usdsek usdnok usddkk usdzar usdhkd
# usdsgd eurpln eurhuf nadchf usdhuf usdpln eurczk
# eursek usdczk zarjpy eurdkk eurnok usddkk

# metals/commodities
# xauusd xauaud xauchf bcousd wtiusd xaueur xagusd xaugbp

# indices
# grxeur auxaud frxeur hkxhkd spxusd jpxjpy udxusd
# nsxusd ukxgbp etxeur

class Pairs(Enum):
    eurusd = "EUR/USD"
    eurchf = "EUR/CHF"
    eurgbp = "EUR/GBP"
    eurjpy = "EUR/JPY"
    euraud = "EUR/AUD"
    usdcad = "USD/CAD"
    usdchf = "USD/CHF"
    usdjpy = "USD/JPY"
    usdmxn = "USD/MXN"
    gbpchf = "GBP/CHF"
    gbpjpy = "GBP/JPY"
    gbpusd = "GBP/USD"
    audjpy = "AUD/JPY"
    audusd = "AUD/USD"
    chfjpy = "CHF/JPY"
    nzdjpy = "NZD/JPY"
    nzdusd = "NZD/USD"
    xauusd = "XAU/USD"
    eurcad = "EUR/CAD"
    audcad = "AUD/CAD"
    cadjpy = "CAD/JPY"
    eurnzd = "EUR/NZD"
    grxeur = "GRX/EUR"
    nzdcad = "NZD/CAD"
    sgdjpy = "SGD/JPY"
    usdhkd = "USD/HKD"
    usdnok = "USD/NOK"
    usdtry = "USD/TRY"
    xauaud = "XAU/AUD"
    audchf = "AUD/CHF"
    auxaud = "AUX/AUD"
    eurhuf = "EUR/HUF"
    eurpln = "EUR/PLN"
    frxeur = "FRX/EUR"
    hkxhkd = "HKX/HKD"
    nadchf = "NZD/CHF"
    spxusd = "SPX/USD"
    usdhuf = "USD/HUF"
    usdpln = "USD/PLN"
    usdzar = "USD/ZAR"
    xauchf = "XAU/CHF"
    zarjpy = "ZAR/JPY"
    bcousd = "BCO/USD"
    etxeur = "ETX/EUR"
    eurczk = "EUR/CZK"
    eursek = "EUR/SEK"
    gbpaud = "GBP/AUD"
    gbpnzd = "GBP/NZD"
    jpxjpy = "JPX/JPY"
    udxusd = "UDX/USD"
    usdczk = "USD/CZK"
    usdsek = "USD/SEK"
    wtiusd = "WTI/USD"
    xaueur = "XAU/EUR"
    audnzd = "AUD/NZD"
    cadchf = "CAD/CHF"
    eurdkk = "EUR/DKK"
    eurnok = "EUR/NOK"
    eurtry = "EUR/TRY"
    gbpcad = "GBP/CAD"
    nsxusd = "NSX/USD"
    ukxgbp = "UKX/GBP"
    usddkk = "USD/DKK"
    usdsgd = "USD/SGD"
    xagusd = "XAG/USD"
    xaugbp = "XAU/GBP"

    @classmethod
    def list_keys(cls):
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls):
        return {member.value for _, member in cls.__members__.items()}


class Format(Enum):
    MT = "metatrader"
    NT = "ninjatrader"
    MS = "metastock"
    ASCII = "ascii"
    XLSX = "excel"

    @classmethod
    def list_keys(cls):
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls):
        return {member.value for _, member in cls.__members__.items()}


class Timeframe(Enum):
    M1 = "1-minute-bar-quotes"
    T = "tick-data-quotes"
    T_LAST = "tick-last-quotes"
    T_BID = "tick-bid-quotes"
    T_ASK = "tick-ask-quotes"

    @classmethod
    def list_keys(cls):
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls):
        return {member.value for _, member in cls.__members__.items()}


class TimeFormat(Enum):
    MT_M1 = "%Y.%m.%d %H:%M"
    ASCII_M1 = "%Y%m%d %H%M%S"
    ASCII_T = "%Y%m%d %H%M%S%f"
    NT_M1 = "%Y%m%d %H%M%S"
    NT_T_LAST = "%Y%m%d %H%M%S"
    NT_T_BID = "%Y%m%d %H%M%S"
    NT_T_ASK = "%Y%m%d %H%M%S"
    MS_M1 = "%Y%m%d%H%M"

    @classmethod
    def list_keys(cls):
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls):
        return {member.value for _, member in cls.__members__.items()}


class TimePrecision(Enum):
    MT_M1 = WritePrecision.S
    ASCII_M1 = WritePrecision.S
    ASCII_T = WritePrecision.MS
    NT_M1 = WritePrecision.S
    NT_T_LAST = WritePrecision.S
    NT_T_BID = WritePrecision.S
    NT_T_ASK = WritePrecision.S
    MS_M1 = WritePrecision.S

    @classmethod
    def list_keys(cls):
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls):
        return {member.value for _, member in cls.__members__.items()}


def get_valid_format_timeframes(csv_format):
    timeframes = []

    match csv_format:
        case "metatrader":
            timeframes.extend(["M1"])
        case "ninjatrader":
            timeframes.extend(["M1", "T_LAST", "T_BID", "T_ASK"])
        case "metastock":
            timeframes.extend(["M1"])
        case "ascii":
            timeframes.extend(["M1", "T"])
        case "excel":
            timeframes.extend(["M1"])

    return timeframes



import os
import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
import multiprocessing
from urllib.parse import urlparse
import requests
from rich import print
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn
import bs4
from bs4 import BeautifulSoup
from histdatacom.utils import get_year_from_datemonth
from histdatacom.utils import get_month_from_datemonth
from histdatacom.utils import get_current_datemonth_gmt_plus5
from histdatacom.fx_enums import Timeframe, get_valid_format_timeframes
from histdatacom.records import Record


class _URLs:
    def __init__(self, args, records_current_, records_next_):
        # setting relationship to global outer parent
        self.args = args

        global records_current
        records_current = records_current_

        global records_next
        records_next = records_next_

        self.args["base_url"] = 'http://www.histdata.com/download-free-forex-data/'
        self.args["post_headers"] = {
            "Host": "www.histdata.com",
            "Connection": "keep-alive",
            "Content-Length": "101",
            "Cache-Control": "max-age=0",
            "Origin": "http://www.histdata.com",
            "Upgrade-Insecure-Requests": "1",
            "DNT": "1",
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "Referer": "",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "en-US,en;q=0.9"}

    def init_counters(self, records_current_, records_next_, args_):
        global records_current
        records_current = records_current_
        global records_next
        records_next = records_next_
        global args
        args = args_

    def populate_initial_queue(self, records_current, records_next):
        for url in self.generate_form_urls(self.args["start_yearmonth"],
                                           self.args["end_yearmonth"],
                                           self.args['formats'],
                                           self.args["pairs"],
                                           self.args['timeframes'],
                                           self.args["base_url"]):
            record = Record()
            record(url=url, status="URL_NEW")
            record.restore_momento(base_dir=self.args['default_download_dir'])

            if record.status != "URL_NO_REPO_DATA":
                record.write_info_file(base_dir=self.args['default_download_dir'])
                records_next.put(record)

        records_next.dump_to_queue(records_current)

    def validate_url(self, record):
        try:
            if record.status == "URL_NEW":
                page_data = self.get_page_data(record.url)
                self.fetch_form_values(page_data, record)

                if record.data_tk == "":
                    raise ValueError

                record.status = "URL_VALID"
                record.write_info_file(base_dir=args['default_download_dir'])

            records_next.put(record)
        except ValueError:
            print(f"Info: Histdata.com does not have: {record.url}")
            record.status = "URL_NO_REPO_DATA"
            record.write_info_file(base_dir=args['default_download_dir'])
        except Exception:
            print(f"Unknown Error for URL: {record.url}", sys.exc_info())
            record.delete_info_file()
            raise
        finally:
            records_current.task_done()

    def validate_urls(self, records_current, records_next):

        records_count = records_current.qsize()
        with Progress(TextColumn(text_format=f"[cyan]Validating {records_count} URLs..."),
                      BarColumn(),
                      "[progress.percentage]{task.percentage:>3.0f}%",
                      TimeElapsedColumn()) as progress:
            task_id = progress.add_task("Validating URLs", total=records_count)

            with ThreadPoolExecutor(max_workers=(multiprocessing.cpu_count() * 3),
                                    initializer=self.init_counters,
                                    initargs=(records_current,
                                              records_next,
                                              self.args.copy())) as executor:
                futures = []

                while not records_current.empty():
                    record = records_current.get()

                    if record is None:
                        return

                    future = executor.submit(self.validate_url, record)
                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    self.complete_future(progress, task_id, futures, future)

        records_current.join()
        records_next.dump_to_queue(records_current)

    def download_zip(self, record):
        try:
            if "URL_VALID" in record.status \
            or (self.args['from_api']
                and not (os.path.exists(record.data_dir + record.zip_filename)
                         or os.path.exists(record.data_dir + record.csv_filename)
                         or os.path.exists(record.data_dir + record.jay_filename))
                ):
                res = self.request_file(record)
                record.zip_filename = res.headers["Content-Disposition"].split(";")[1].split("=")[1]
                self.write_file(record, res.content)

                record.status = "CSV_ZIP" if record.status == "URL_VALID" else record.status
                record.write_info_file(base_dir=args['default_download_dir'])

            records_next.put(record)
        except KeyError:
            print(f"Invalid Zip on histdata.com: {record.url}", sys.exc_info())
            record.delete_info_file()
        except Exception:
            print("Unexpected error:", sys.exc_info())
            record.delete_info_file()
            raise
        finally:
            records_current.task_done()

    def request_file(self, record):
        post_headers = args['post_headers'].copy()
        post_headers["Referer"] = record.url
        return requests.post("http://www.histdata.com/get.php",
                             data={"tk": record.data_tk,
                                   "date": record.data_date,
                                   "datemonth": record.data_datemonth,
                                   "platform": record.data_format,
                                   "timeframe": record.data_timeframe,
                                   "fxpair": record.data_fxpair},
                             headers=post_headers)

    def write_file(self, record, content):
        zip_path = record.data_dir + record.zip_filename
        with open(zip_path, "wb") as zip_file:
            zip_file.write(content)

    def download_zips(self, records_current, records_next):

        records_count = records_current.qsize()
        with Progress(TextColumn(text_format=f"[cyan]Downloading {records_count} ZIPs..."),
                      BarColumn(),
                      "[progress.percentage]{task.percentage:>3.0f}%",
                      TimeElapsedColumn()) as progress:
            task_id = progress.add_task("[cyan]Downloading ZIPs", total=records_count)

            with ThreadPoolExecutor(max_workers=(multiprocessing.cpu_count() * 3),
                                    initializer=self.init_counters,
                                    initargs=(records_current,
                                              records_next,
                                              self.args.copy())) as executor:
                futures = []

                while not records_current.empty():
                    record = records_current.get()

                    if record is None:
                        return

                    future = executor.submit(self.download_zip, record)
                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    self.complete_future(progress, task_id, futures, future)
        records_current.join()

        records_next.dump_to_queue(records_current)

    def complete_future(self, progress, task_id, futures, future):
        progress.advance(task_id, 0.75)
        futures.remove(future)
        del future

    @classmethod
    def get_page_data(cls, url):
        request = requests.get(url)

        page_content = BeautifulSoup(request.content, "html.parser")
        encoding = dict(request.headers)['Content-Encoding']
        bytes_length = dict(request.headers)['Content-Length']

        return {"page_content": page_content, "encoding": encoding, "bytes_length": bytes_length}

    @classmethod
    def get_base_url(cls, url):
        parsed_url = urlparse(url)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @classmethod
    def fetch_form_values(cls, page_data, record):
        form_page_content = page_data['page_content']
        form = form_page_content.find_all('form', id='file_down')

        for element in form:
            for value in element:
                if type(value) is bs4.element.Tag:
                    match value.get("id"):
                        case "tk":
                            record.data_tk = value.get("value")
                        case "date":
                            record.data_date = value.get("value")
                        case "datemonth":
                            record.data_datemonth = value.get("value")
                        case "platform":
                            record.data_format = value.get("value")
                        case "timeframe":
                            record.data_timeframe = value.get("value")
                        case "fxpair":
                            record.data_fxpair = value.get("value")
        return record

    @classmethod
    def valid_format_timeframe_pair_urls(cls, formats, timeframes, pairs):
        for csv_format in formats:
            for timeframe in timeframes:
                if timeframe in get_valid_format_timeframes(csv_format):
                    for pair in pairs:
                        yield f"{csv_format}/{Timeframe[timeframe].value}/{pair}/", timeframe

    @classmethod
    def correct_for_zero_month(cls, month):
        if month == 0:
            month = 1
        return month

    @classmethod
    def generate_form_urls(cls,
                           start_yearmonth,
                           end_yearmonth,
                           formats,
                           pairs,
                           timeframes,
                           base_url):
        current_yearmonth = get_current_datemonth_gmt_plus5()
        current_year = int(get_year_from_datemonth(current_yearmonth))

        if start_yearmonth is None and end_yearmonth is None:
            start_yearmonth, end_yearmonth = "200001", current_yearmonth

        for sub_url, timeframe in cls.valid_format_timeframe_pair_urls(formats, timeframes, pairs):
            form_url = f"{base_url}?/{sub_url}"

            if end_yearmonth is None:
                for date_url in cls.yield_single_year_or_month(timeframe, start_yearmonth):
                    yield f"{form_url}{date_url}"
            else:
                start_year = int(get_year_from_datemonth(start_yearmonth))
                start_month = int(get_month_from_datemonth(start_yearmonth))
                end_year = int(get_year_from_datemonth(end_yearmonth))
                end_month = int(get_month_from_datemonth(end_yearmonth))

                for year in range(start_year, end_year + 1):
                    yield from cls.yield_range_of_yearmonths(year, timeframe, form_url,
                                                             start_year, start_month,
                                                             end_year, end_month,
                                                             current_year)

    @classmethod
    def yield_range_of_yearmonths(cls, year, timeframe, form_url,
                                  start_year, start_month,
                                  end_year, end_month,
                                  current_year):

        match year:
            case _ if year == current_year:
                for date_url in cls.yield_current_year(year,
                                                       start_year,
                                                       start_month,
                                                       end_year,
                                                       end_month):
                    yield f"{form_url}{date_url}"

            case _ if start_year == year == end_year:
                for date_url in cls.yield_same_year(timeframe, year, start_month, end_month):
                    yield f"{form_url}{date_url}"

            case _ if year == start_year != end_year:
                for date_url in cls.yield_start_year(timeframe, year, start_month):
                    yield f"{form_url}{date_url}"

            case _ if year == end_year != start_year:
                for date_url in cls.yield_end_year(timeframe, year, end_month):
                    yield f"{form_url}{date_url}"
            case _:
                for date_url in cls.yield_year(timeframe, year):
                    yield f"{form_url}{date_url}"

    @classmethod
    def yield_current_year(cls, year, start_year, start_month, end_year, end_month):
        if start_year == end_year:
            for month in range(start_month, end_month + 1):
                yield f"{year}/{month}"
        else:
            for month in range(1, end_month + 1):
                yield f"{year}/{month}"

    @classmethod
    def yield_same_year(cls, timeframe, year, start_month, end_month):
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(start_month, end_month + 1):
                    yield f"{year}/{month}"

    @classmethod
    def yield_start_year(cls, timeframe, year, start_month):
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(start_month, 12 + 1):
                    yield f"{year}/{month}"

    @classmethod
    def yield_end_year(cls, timeframe, year, end_month):
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(1, end_month + 1):
                    yield f"{year}/{month}"

    @classmethod
    def yield_year(cls, timeframe, year):
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(1, 12 + 1):
                    yield f"{year}/{month}"

    @classmethod
    def yield_single_year_or_month(cls, timeframe, start_yearmonth):

        current_yearmonth = get_current_datemonth_gmt_plus5()
        current_year = int(get_year_from_datemonth(current_yearmonth))
        current_month = int(get_month_from_datemonth(current_yearmonth))

        start_year = int(get_year_from_datemonth(start_yearmonth))
        start_month = int(get_month_from_datemonth(start_yearmonth))

        if start_month == 0:  # return the year's data
            if start_year == current_year:
                for month in range(1, current_month + 1):
                    yield f"{start_year}/{month}"
            else:
                for date_url in cls.yield_year(timeframe, start_year):
                    yield date_url
        else:
            if start_year == current_year:
                yield f"{start_year}/{start_month}"
            else:
                match timeframe:
                    case "M1":
                        yield f"{start_year}"
                    case _:
                        yield f"{start_year}/{start_month}"



from histdatacom.fx_enums import Pairs
from histdatacom.fx_enums import Format
from histdatacom.fx_enums import Timeframe

class Options:
    """ An intra-class DTO for Default Arguments for _HistDataCom class. """
    # argparse uses a thin class to create a namespace for cli/shell arguments to live in
    # normally argparse.ArgumentParser.parse_args(namespace=...) creates this namespace and
    # writes user's cli args to it.  Preemptively creating here to hold default args; if the
    # user enters args in the shell, these values will be respectively overwritten
    def __init__(self):
        self.validate_urls = False
        self.download_data_archives = False
        self.extract_csvs = False
        self.import_to_influxdb = False
        self.pairs = Pairs.list_keys()
        self.formats = Format.list_values()
        self.timeframes = Timeframe.list_keys()
        self.start_yearmonth = ""
        self.end_yearmonth = ""
        self.data_directory = "data"
        self.from_api = False
        self.api_return_type = "datatable"


import os
import sys
import csv
import re
from datetime import datetime
import pytz
import yaml
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn


def get_month_from_datemonth(datemonth):
    return datemonth[-2:] if datemonth is not None and len(datemonth) > 4 else ""


def get_year_from_datemonth(datemonth):
    return datemonth[:4] if datemonth is not None else ""


def get_query_string(url):
    return url.split('?')[1].split('/')


def create_full_path(path_str):
    if not os.path.exists(path_str):
        os.makedirs(path_str)


def set_working_data_dir(data_dirname):
    return f"{os.getcwd()}{os.sep}{data_dirname}{os.sep}"


def load_influx_yaml():

    if os.path.exists('influxdb.yaml'):
        with open('influxdb.yaml', 'r') as file:
            try:
                yamlfile = yaml.safe_load(file)
            except yaml.YAMLError as exc:
                print(exc)
                sys.exit()
        return yamlfile

    print(""" ERROR: -I flag is used to import data to a influxdb instance...
                        there is no influxdb.yaml file in working directory.
                        did you forget to set it up?
          """)
    sys.exit()


def get_current_datemonth_gmt_plus5():
    now = datetime.now().astimezone()
    gmt_plus5 = now.astimezone(pytz.timezone("Etc/GMT+5"))
    return f"{gmt_plus5.year}{gmt_plus5.strftime('%m')}"


def get_progress_bar(progress_string):

    return \
        TextColumn(text_format=progress_string), \
        BarColumn(),\
        "[progress.percentage]{task.percentage:>3.0f}%", \
        TimeElapsedColumn()


def get_csv_dialect(csv_path):
    with open(csv_path, "r") as srccsv:
        dialect = csv.Sniffer().sniff(srccsv.read(), delimiters=",; ")
    return dialect


def replace_date_punct(datemonth_str):
    """removes year-month punctuation and returns str("000000")"""
    return re.sub("[-_.: ]", "", datemonth_str) if datemonth_str is not None else ""


import sys
from multiprocessing import managers
from histdatacom.cli import ArgParser
from histdatacom.options import Options
from histdatacom.records import Records
from histdatacom.utils import set_working_data_dir
from histdatacom.utils import load_influx_yaml
from histdatacom.urls import _URLs
from histdatacom.api import _API
from histdatacom.csvs import _CSVs
from histdatacom.influx import _Influx


class _HistDataCom:
    """A module to pull market data from histdata.com and import it into influxDB"""

    def __init__(self,
                 records_current,
                 records_next,
                 csv_chunks_queue,
                 options):

        """ Initialization for _HistDataCom Class"""
        # Set User () or Default Arguments respectively utilizing the self.ArgParser
        # and self.Options classes.
        #   - ArgParser()():
        #       - ()(): use an IIFE to allow argparse to get garbage collected
        #       - ()(): ArgParser.__call__ returns updated Options object
        #       - vars(...): get the __dict__ representation of the object
        #       - ArgParser._arg_list_to_set(...)
        #           - Normalize iterable user arguments whose values are lists and
        #             make them sets instead
        #       - .copy(): decouple for GC using a hard copy of user args
        self.args = ArgParser._arg_list_to_set(vars(ArgParser(options)())).copy()
        self.args['default_download_dir'] = set_working_data_dir(self.args['data_directory'])
        self.args['queue_filename'] = ".queue"

        if self.args["import_to_influxdb"] == 1:
            influx_yaml = load_influx_yaml()
            self.args['INFLUX_ORG'] = influx_yaml['influxdb']['org']
            self.args['INFLUX_BUCKET'] = influx_yaml['influxdb']['bucket']
            self.args['INFLUX_URL'] = influx_yaml['influxdb']['url']
            self.args['INFLUX_TOKEN'] = influx_yaml['influxdb']['token']

        self.records_current = records_current
        self.records_next = records_next

        self.urls = _URLs(self.args, self.records_current, self.records_next)
        self.csvs = _CSVs(self.args, self.records_current, self.records_next)
        self.api = _API(self.args, self.records_current, self.records_next)

        if self.args["import_to_influxdb"] == 1:
            self.csv_chunks_queue = csv_chunks_queue
            self.influx = _Influx(self.args,
                                  self.records_current,
                                  self.records_next,
                                  self.csv_chunks_queue)

    def run(self):
        self.urls.populate_initial_queue(self.records_current, self.records_next)

        if self.args["validate_urls"]:
            self.urls.validate_urls(self.records_current, self.records_next)

        if self.args["download_data_archives"]:
            self.urls.download_zips(self.records_current, self.records_next)
            if self.args["from_api"]:
                self.api.validate_jays(self.records_current, self.records_next)
                return self.api.merge_jays(self.records_current, self.records_next)

        if self.args["from_api"]:
            print("return data to api call from histdata_com.run()")

        if self.args["extract_csvs"]:
            self.csvs.extract_csvs(self.records_current, self.records_next)

        if self.args["import_to_influxdb"]:
            self.influx.import_data(self.records_current,
                                    self.records_next,
                                    self.csv_chunks_queue)


class HistDataCom():
    def __init__(self, options):
        self.options = options
        self.records_manager=managers.SyncManager()
        self.records_manager.register("Records", Records)

    def __call__(self):
        self.records_manager.start()

        global records_current
        records_current = self.records_manager.Records()

        global records_next
        records_next = self.records_manager.Records()

        global csv_chunks_queue
        csv_chunks_queue = self.records_manager.Queue()

        scraper = _HistDataCom(records_current,
                                records_next,
                                csv_chunks_queue,
                                self.options)

        if self.options.from_api:
            return scraper.run()
        else:
            scraper.run()
          
def main(options=None):
    if not options:
        options = Options()
        HistDataCom(options)()
    else:
        options.from_api = True
        return HistDataCom(options)()



if __name__ == '__main__':
    sys.exit(main())



from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
import multiprocessing
import sys
import os
import zipfile
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn
from rich import print


class _CSVs:
    def __init__(self, args_, records_current_, records_next_):
        # setting relationship to global outer parent
        self.args = args_

        global records_current
        records_current = records_current_

        global records_next
        records_next = records_next_

    def init_counters(self, records_current_, records_next_, args_):
        global records_current
        records_current = records_current_
        global records_next
        records_next = records_next_
        global args
        args = args_

    def extract_csv(self, record):
        try:
            if "CSV_ZIP" in record.status:
                zip_path = record.data_dir + record.zip_filename

                with zipfile.ZipFile(zip_path, "r") as zip_ref:
                    [record.csv_filename] = [x for x in zip_ref.namelist() if (".csv" or ".xlsx") in x]
                    zip_ref.extract(record.csv_filename, path=record.data_dir)

                os.remove(zip_path)
                record.status = "CSV_FILE"
                record.write_info_file(base_dir=args['default_download_dir'])
            records_next.put(record)
        except Exception:
            print("Unexpected error:", sys.exc_info())
            record.delete_info_file()
            raise
        finally:
            records_current.task_done()

    def extract_csvs(self, records_current, records_next):

        records_count = records_current.qsize()
        with Progress(TextColumn(text_format=f"[cyan]Extracting {records_count} CSVs..."),
                      BarColumn(),
                      "[progress.percentage]{task.percentage:>3.0f}%",
                      TimeElapsedColumn()) as progress:

            task_id = progress.add_task("[cyan]Extracting CSVs", total=records_count)

            with ProcessPoolExecutor(max_workers=(multiprocessing.cpu_count() - 1),
                                     initializer=self.init_counters,
                                     initargs=(records_current,
                                     records_next,
                                     self.args.copy())) as executor:
                futures = []

                while not records_current.empty():
                    record = records_current.get()

                    if record is None:
                        return

                    future = executor.submit(self.extract_csv, record)
                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    progress.advance(task_id, 0.75)
                    futures.remove(future)
                    del future

        records_current.join()
        records_next.dump_to_queue(records_current)



import codecs
import os.path
from setuptools import setup
from setuptools import find_packages

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

setup(
    # basic package data
    name="histdatacom",
    version=get_version("src/histdatacom/__init__.py"),
    description="A Multi-threaded/Multi-Process command-line utility and python package that downloads currency exchange rates from Histdata.com. Imports to InfluxDB. Can be used in Jupyter Notebooks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/dmidlo/histdata.com-tools',
    project_urls={
        "Bug Tracker": "https://github.com/dmidlo/histdata.com-tools/issues",
    },
    author="David Midlo",
    author_email="dmidlo@gmail.com",
    license="MIT License",

    # package structure
    packages=find_packages('src'),
    package_dir={'': 'src'},

    # install the RSReader executable
    entry_points={
        'console_scripts': [
            'histdatacom = histdatacom.histdata_com:main'
        ]
    },
    install_requires=[
        'influxdb_client',
        'rich',
        'requests',
        'bs4',
        'pyyaml',
        'rx',
        'argparse',
        'pytz',
        'ipywidgets'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Environment :: MacOS X',
        'Operating System :: MacOS',
        'Environment :: Win32 (MS Windows)',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: Jupyter',
        'Topic :: Terminals',
    ]
)

#print("\n\n\n done............. pip install git+https://github.com/h2oai/datatable")


import os
import sys
import io
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
from urllib.request import urlopen
from csv import DictReader
from functools import partial
from datetime import datetime
import yaml
import rx
import pytz
from rx import operators as ops
from influxdb_client import InfluxDBClient
from influxdb_client import WriteOptions
from influxdb_client import WritePrecision
from influxdb_client.client.write_api import WriteType
from rich.progress import Progress
from rich.progress import BarColumn, TextColumn, TimeElapsedColumn, SpinnerColumn
from rich import print
from histdatacom.fx_enums import TimeFormat
from histdatacom.utils import get_csv_dialect


class _InfluxDBWriter(multiprocessing.Process):
    def __init__(self, args, csv_chunks_queue):
        multiprocessing.Process.__init__(self)
        self.args = args
        self.csv_chunks_queue = csv_chunks_queue
        self.client = InfluxDBClient(url=self.args['INFLUX_URL'],
                                     token=self.args['INFLUX_TOKEN'],
                                     org=self.args['INFLUX_ORG'],
                                     debug=False)
        self.write_api = self.client.write_api(write_options=WriteOptions(
                                               write_type=WriteType.batching,
                                               batch_size=25_000,
                                               flush_interval=12_000))

    def run(self):
        while True:
            chunk = self.csv_chunks_queue.get()

            if chunk is None:
                self.terminate()
                self.csv_chunks_queue.task_done()
                break

            self.write_api.write(org=self.args['INFLUX_ORG'],
                                 bucket=self.args['INFLUX_BUCKET'],
                                 record=chunk,
                                 write_precision=WritePrecision.MS)
            self.csv_chunks_queue.task_done()

    def terminate(self):
        self.write_api.__del__()
        self.client.__del__()


class _Influx():
    def __init__(self, args, records_current_, records_next_, csv_chunks_queue_):
        self.args = args

        global records_current
        records_current = records_current_

        global records_next
        records_next = records_next_

        global csv_chunks_queue
        csv_chunks_queue = csv_chunks_queue_

    def init_counters(self, csv_chunks_queue_, records_current_, records_next_, args_):
        global csv_chunks_queue
        csv_chunks_queue = csv_chunks_queue_
        global records_current
        records_current = records_current_
        global records_next
        records_next = records_next_
        global args
        args = args_

    def parse_row(self, row, record):
        # line protocol example: myMeasurement,tag1=value1,tag2=value2 fieldKey="fieldValue" 1556813561098000000
        measurement = f"{record.data_fxpair}"
        tags = f"source=histdata.com,format={record.data_format},timeframe={record.data_timeframe}".replace(" ", "")
        time = self.convert_datetime_to_utc_timestamp(record.data_format,
                                                      record.data_timeframe,
                                                      row)

        match record.data_timeframe:
            case "M1":
                fields = f"openbid={row['openBid']},highbid={row['highBid']},lowbid={row['lowBid']},closebid={row['closeBid']}".replace(" ", "")
            case "T":
                fields = f"bidquote={row['bidQuote']},askquote={row['askQuote']}".replace(" ", "")

        line_protocol = f"{measurement},{tags} {fields} {time}"

        return line_protocol

    def parse_rows(self, rows, record):
        mapfunc = partial(self.parse_row, record=record)
        _parsed_rows = list(map(mapfunc, rows))

        csv_chunks_queue.put(_parsed_rows)

    def import_file(self, record):
        try:
            if ("CSV" in record.status) \
            and ("ZIP" not in record.status) \
            and str.lower(record.data_format) == "ascii":
                self.import_csv(record)
            records_next.put(record)
        except Exception:
            print("Unexpected error from here:", sys.exc_info())
            record.delete_into_file()
            raise
        finally:
            records_current.task_done()

    def import_csv(self, record):
        csv_path = record.data_dir + record.csv_filename
        file_endpoint = f"file://{record.data_dir}{record.csv_filename}"

        res = urlopen(file_endpoint)
        io_wrapper = io.TextIOWrapper(res)

        with ProcessPoolExecutor(max_workers=(1 + (multiprocessing.cpu_count() // 2)),
                                 initializer=self.init_counters,
                                 initargs=(csv_chunks_queue,
                                           records_current,
                                           records_next,
                                           self.args.copy())) as executor:

            fieldnames = self.fieldnames_match(record.data_format, record.data_timeframe)
            dialect = get_csv_dialect(csv_path)
            data = rx.from_iterable(
                DictReader(io_wrapper,
                           fieldnames=fieldnames,
                           dialect=dialect)) \
                .pipe(
                    ops.buffer_with_count(25_000),
                    ops.flat_map(
                        lambda rows: executor.submit(self.parse_rows, rows, record)))

            data.subscribe(
                on_next=lambda x: None,
                on_error=lambda er: print(f"Unexpected error: {er}"))

        os.remove(csv_path)
        record.status = "INFLUX_UPLOAD"
        record.write_info_file(base_dir=args['default_download_dir'])

    def import_data(self, records_current, records_next, csv_chunks_queue):

        writer = _InfluxDBWriter(self.args, csv_chunks_queue)
        writer.start()

        records_count = records_current.qsize()
        with Progress(TextColumn(text_format=f"[cyan]Adding {records_count} CSVs to influx queue..."),
                      BarColumn(),
                      "[progress.percentage]{task.percentage:>3.0f}%",
                      TimeElapsedColumn()) as progress:
            task_id = progress.add_task("influx", total=records_count)

            # cpu count -1: Manager, -1: DBWriter,
            with ProcessPoolExecutor(max_workers=(1 + (multiprocessing.cpu_count() // 2)),
                                     initializer=self.init_counters,
                                     initargs=(csv_chunks_queue,
                                               records_current,
                                               records_next,
                                               self.args.copy())) as executor:
                futures = []

                while not records_current.empty():
                    record = records_current.get()

                    if record is None:
                        break

                    future = executor.submit(self.import_file, record)
                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    progress.advance(task_id, 0.75)
                    futures.remove(future)
                    del future

        with Progress(TextColumn(text_format="[cyan]...finishing upload to influxdb"),
                      SpinnerColumn(), SpinnerColumn(), SpinnerColumn(),
                      TimeElapsedColumn()) as progress:
            task_id = progress.add_task("waiting", total=0)

            records_current.join()
            csv_chunks_queue.put(None)
            csv_chunks_queue.join()
            progress.advance(task_id, 0.75)

        print("[cyan] done.")
        records_next.dump_to_queue(records_current)

    @classmethod
    def load_influx_yaml(cls):

        if os.path.exists('influxdb.yaml'):
            with open('influxdb.yaml', 'r') as file:
                try:
                    yamlfile = yaml.safe_load(file)
                except yaml.YAMLError as exc:
                    print(exc)
                    sys.exit()

            return yamlfile
        
        print(""" ERROR: -I flag is used to import data to a influxdb instance...
                          there is no influxdb.yaml file in working directory.
                          did you forget to set it up?
              """)
        sys.exit()

    @classmethod
    def fieldnames_match(cls, csv_format, timeframe):
        try:
            match csv_format:
                case "ASCII" if timeframe == "M1":
                    fieldnames = ["msSinceEpochUTC", "openBid", "highBid", "lowBid", "closeBid", "Volume"]
                case "ASCII" if timeframe == "T":
                    fieldnames = ["msSinceEpochUTC", "bidQuote", "askQuote", "Volume"]
                case _:
                    raise ValueError("Invalid format for influx import")
            return fieldnames
        except ValueError as err:
            print(err)
            sys.exit()

    @classmethod
    def get_timeformat(cls, csv_format, timeframe):

        format_enum_key = f'{str(csv_format)}_{str(timeframe)}'

        return TimeFormat[format_enum_key].value

    @classmethod
    def convert_datetime_to_utc_timestamp(cls, csv_format, timeframe, row):

        est_timestamp = row["msSinceEpochUTC"]
        date_object = datetime.strptime(est_timestamp, cls.get_timeformat(csv_format, timeframe))
        tz_date_object = date_object.replace(tzinfo=pytz.timezone("Etc/GMT+5"))

        timestamp = int(tz_date_object.timestamp() * 1000)

        return str(timestamp)
